"""Opencv Image Utilities."""
import uuid

import cv2


def load_image(image_filename):
    """Load image with OpenCV and colored i.e. BGR.

    Args:
        image_filename (str): Path of image.

    Returns:
        numpy.ndarray: Image array with the following characteristics
        - pixel_values: [0, 255]
        - channel_order: BGR
        - shape_order: HWC
        - dtype: np.uint8
    """
    return cv2.imread(image_filename)


def show_image(image, window_name=None, wait=True, wait_time=0, destroy=True, destroy_all_windows=True):
    """Display an image on a window and wait.

    Args:
        image (numpy.ndarray): Image array to display.
        window_name (str): Name of window which is used for identification while destroying.
            If None, a random window_name is created on each call which also generates a new window.
        wait (bool): Whether to wait after displaying the image.
        wait_time (int): If `wait` is True, then the `wait_time` indicates the
            time in milliseconds to wait after displaying the image.
        destroy (bool): Whether to destroy the window after displaying the image.
        destroy_all_windows (bool): Whether to destroy all windows of just the
            window that was created by this function call.
            If True, destroys all windows by calling `cv2.destroyAllWindows()`.
            If False, destroys just this function call window by calling `cv2.destroyWindow(window_name)`.

    """
    window_name = window_name if window_name else "window_{}".format(str(uuid.uuid4())[:8])
    already_destroyed = False

    cv2.imshow(window_name, image)

    if wait:
        # It should be possible to destroy the window while waiting.
        # Press q to destroy.
        if cv2.waitKey(wait_time) & 0xFF == ord('q'):
            if destroy_all_windows:
                cv2.destroyAllWindows()
            else:
                cv2.destroyWindow(window_name)
            already_destroyed = True

    # If window is already destroyed, don't destroy again!
    if destroy and not already_destroyed:
        if destroy_all_windows:
            cv2.destroyAllWindows()
        else:
            cv2.destroyWindow(window_name)



"""."""
import logging

from flask import Blueprint, jsonify, render_template, request

from . import db_api, web_api

logger = logging.getLogger(__name__)

job_status_handler = Blueprint(name='job_status',
                               import_name=__name__,
                               template_folder='templates',
                               static_folder='static')


# ##############################################################################
# ############################ PAGES ###########################################
# ##############################################################################
@job_status_handler.route('/', methods=['GET'])
def index():
    """Index."""
    return render_template('js_index.html')


@job_status_handler.route('/status/<page_id>', methods=['GET', 'POST'])
def status(page_id):
    """Render draw status page."""
    def get_is_reload(jobs):
        for job in jobs:
            if job.status != 'SUCCESS':
                return True
        # If all statuses are SUCCESS, then don't reload anymore.
        return False

    if page_id == "all":
        jobs = web_api.read_all_jobs()
        is_reload = str(get_is_reload(jobs))
        return render_template('js_status.html', jobs=jobs, page_id='all', is_reload=is_reload)
    else:
        job = web_api.read_job(job_id=page_id)
        jobs = [job]
        is_reload = str(get_is_reload(jobs))
        return render_template('js_status.html', jobs=jobs, page_id=page_id, is_reload=is_reload)


# ##############################################################################
# ############################ CREATE ##########################################
# ##############################################################################


@job_status_handler.route('/create_job', methods=['POST'])
def create_job():
    """Create new task."""
    name = request.form.get('name', None)
    uuid = request.form.get('uuid', None)
    logger.debug('/create_job: name: {}, uuid: {}'.format(name, uuid))

    job = db_api.insert_new_job(uuid=uuid,
                                name=name,
                                payload_text=request.form.get('payload_text', None),
                                status=request.form.get('status', None))
    return jsonify(job.as_json())


@job_status_handler.route('/create_task', methods=['POST'])
def create_task():
    """Create task."""
    job_id = request.form.get('job_id', None)
    logger.debug('/create_task: job_id: {}'.format(job_id))

    task = db_api.insert_new_task(job_id=job_id,
                                  uuid=request.form.get('uuid', None),
                                  name=request.form.get('name', None),
                                  payload_text=request.form.get('payload_text', None),
                                  status=request.form.get('status', None))
    return jsonify(task.as_json())


# ##############################################################################
# ############################# UPDATE #########################################
# ##############################################################################
@job_status_handler.route('/update_job', methods=['POST'])
def update_job():
    """Update job."""
    job_id = request.form.get('job_id', None)
    uuid = request.form.get('uuid', None)
    name = request.form.get('name', None)
    payload_text = request.form.get('payload_text', None)
    status = request.form.get('status', None)
    logger.debug('/update_job: job_id: {}, uuid: {}, name: {}, status: {}, payload_text: {}'
                 .format(job_id, uuid, name, status, payload_text))

    job = db_api.update_job(job_id=job_id,
                            uuid=uuid,
                            name=name,
                            payload_text=payload_text,
                            status=status)
    return jsonify(job.as_json())


@job_status_handler.route('/update_task', methods=['POST'])
def update_task():
    """Update task."""
    task_id = request.form.get('task_id', None)
    uuid = request.form.get('uuid', None)
    name = request.form.get('name', None)
    payload_text = request.form.get('payload_text', None)
    status = request.form.get('status', None)
    update_parent_job = request.form.get('update_parent_job', None)
    logger.debug('/update_task: task_id: {}, uuid: {}, name: {}, status: {}, payload_text: {}, update_parent_job: {}'
                 .format(task_id, uuid, name, status, payload_text, update_parent_job))

    task = db_api.update_task(task_id=task_id,
                              uuid=uuid,
                              name=name,
                              payload_text=payload_text,
                              status=status,
                              update_parent_job=update_parent_job)
    return jsonify(task.as_json())


# ##############################################################################
# ############################# READ ###########################################
# ##############################################################################
@job_status_handler.route('/read_job', methods=['GET'])
def read_job():
    """Read job."""
    job_id = request.args['job_id']
    logger.debug('/read_job: job_id: {}'.format(job_id))

    job = db_api.select_job_from_id(job_id)
    return jsonify(job.as_json())


@job_status_handler.route('/read_all_jobs', methods=['GET'])
def read_all_jobs():
    """Read all jobs."""
    logger.debug('/read_all_jobs')

    jobs = db_api.select_job_all()
    jobs = [job.as_json() for job in jobs]
    return jsonify(jobs)


@job_status_handler.route('/read_task', methods=['GET'])
def read_task():
    """Read task."""
    task_id = request.args.get('task_id', None)
    logger.debug('/read_task: task_id: {}'.format(task_id))

    if task_id is None:
        raise ValueError('task_id is None at /read_task')
    task = db_api.select_task_from_id(task_id)
    return jsonify(task.as_json())


@job_status_handler.route('/read_task_from_uuid', methods=['GET'])
def read_task_from_uuid():
    """Read task from uuid."""
    task_uuid = request.args.get('task_uuid', None)
    logger.debug('/read_task: task_uuid: {}'.format(task_uuid))

    if task_uuid is None:
        raise ValueError('task_uuid is None at /read_task')
    task = db_api.select_task_from_uuid(task_uuid)
    return jsonify(task.as_json())



from .image_utils_cv2 import load_image as load_image_cv2
from .image_utils_cv2 import show_image as show_image_cv2
from .image_utils_meta import load_image as load_image_meta
from .image_utils_meta import transform_image as transform_image_meta



"""Utils."""
import logging

logger = logging.getLogger(__name__)


class DotMap(dict):
    """DotMap from a dictionary.

    Example:
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """

    def __init__(self, *args, **kwargs):
        """."""
        super().__init__(*args, **kwargs)

        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super().__delitem__(key)
        del self.__dict__[key]

    def as_json(self):
        """Return json/dict of key-value pairs."""
        return self.__dict__


class DotMapJob(DotMap):
    """DotMap Job which contains DotMap tasks."""

    def __init__(self, *args, **kwargs):
        """."""
        super().__init__(*args, **kwargs)

        if len(self.tasks) != 0:
            self.tasks = [DotMap(task_dict) for task_dict in self.tasks]


class DotMapTask(DotMap):
    """DotMap Task which is just an alias for DotMap."""
    pass



"""."""
import logging
import math
from collections import namedtuple

import cv2
import numpy as np

import nr_common.image_utils.image_utils as image_utils
import nr_common.image_utils.image_utils_cv2 as cv2_utils
import nr_common.image_utils.image_utils_pil as pil_utils

logger = logging.getLogger(__name__)

MetaType = namedtuple(typename="Meta", field_names="scale channel_order shape_order dtype")
META_TYPE_CV2 = MetaType(scale=255, channel_order="BGR", shape_order="HWC", dtype=np.uint8)
META_TYPE_CAFFE = MetaType(scale=1, channel_order="RGB", shape_order="HWC", dtype=np.float32)

ImageWithMeta = namedtuple(typename="ImageWithMeta", field_names="image scale channel_order shape_order dtype")
ImageWithMeta.__doc__ += """.

A namedtuple to store an image array (numpy.ndarray) along with some meta information:
scale (int): The scale of the pixel values.
    - `1` would mean pixel value range of [0-1]
    - `255` would mean pixel value range of [0-255]
channel_order (str): The order of the channels. Currently supported channel_orders are:
    - "RGB"
    - "BGR"
shape_order (str): The order of the axes of the image array. Currently supported shape_orders are:
    - "HWC"
    - "CHW"
dtype (type): The type of the output array. Can be numpy types such as np.float32, np.uint8 or
    standard types such as float or int.
"""


def add_meta_to_image(image, scale=None, channel_order=None, shape_order=None, dtype=None, meta_type=None):
    """."""
    if meta_type:
        scale = meta_type.scale
        channel_order = meta_type.channel_order
        shape_order = meta_type.shape_order
        dtype = meta_type.dtype

    return ImageWithMeta(image=image, scale=scale, channel_order=channel_order, shape_order=shape_order, dtype=dtype)


def load_image(image_filename, scale=255, channel_order="RGB", shape_order="HWC", dtype=np.float32,
               with_meta=False, max_area=None):
    """Load image from filename and transform it to have some pre-defined properties.

    Args:
        image_filename (str): The full path of the image file. The image is expected to be 3D.
        scale (int): The scale of the pixel values.
            - `1` would mean pixel value range of [0-1]
            - `255` would mean pixel value range of [0-255]
        channel_order (str): The order of the channels. Currently supported channel_orders are:
            - "RGB"
            - "BGR"
        shape_order (str): The order of the axes of the image array. Currently supported shape_orders are:
            - "HWC"
            - "CHW"
        with_meta (bool): Whether a simple numpy.ndarray is returned or a namedtuple that contains the image
            and it's meta info is returned.
            - If True, image array with meta-info (ImageWithMeta) is returned
            - If False, image array (numpy.ndarray) is returned
        dtype (type): The type of the output array. Can be numpy types such as np.float32, np.uint8 or
            standard types such as float or int.
        max_area (int): The maximum area i.e. maximum total pixels allowed in the image. If the loaded
            area is greater than max_area then the image is resized keeping the aspect ratio.

    Returns:
        numpy.ndarray: A 3D image array that has the pre-defined properties.
    """
    logger.debug("LOAD IMAGE: scale {} ::: channel_order {} ::: shape_order {} ::: with_meta {} ::: filename {}"
                 .format(scale, channel_order, shape_order, with_meta, image_filename))
    if channel_order == "RGB":
        image = pil_utils.load_image(image_filename)  # RGB, 255, HWC
    elif channel_order == "BGR":
        image = cv2_utils.imread(image_filename)      # BGR, 255, HWC

    # NOTE: This is to make sure that the loaded images doesn't have an
    # area (height * width) greater than max_area.
    # NOTE: This method expects HWC channels & np.uint8 dtype
    if max_area:
        image = image_utils.resize_image_if_exceeds_max_area(image, max_area)

    if scale == 1:
        # The assumption is that image will always be scaled at 255 when it reaches here since all libraries
        # i.e. PIL and OpenCV read images from files at 255 scale.
        image = image / 255.0

    if shape_order == "CHW":
        # TODO: Transform
        raise NotImplementedError

    if image.dtype != dtype:
        image = image.astype(dtype)

    if with_meta is True:
        image = ImageWithMeta(image=image, scale=scale, channel_order=channel_order,
                              shape_order=shape_order, dtype=dtype)

    return image


# TODO: Remove image param
def transform_image(image=None, image_with_meta=None, scale=None, channel_order=None,
                    shape_order=None, dtype=None, meta_type=None, with_meta=False):
    """Transform image or image_with_meta based on some pre-definied properties.

    Args:
        image (numpy.ndarray, ImageWithMeta): ImageWithMeta object that contains the image array and some meta information
            that is to be transformed.
        scale (int): The scale of the pixel values of transformed image.
            - `1` would mean pixel value range of [0-1]
            - `255` would mean pixel value range of [0-255]
        channel_order (str): The order of the channels of the transformed image.
            Currently supported channel_orders are:
            - "RGB" (ImageWithMeta)
            - "BGR" (ImageWithMeta)
            - "REVERSE_CHANNELS" (numpy.ndarray)
        shape_order (str): The order of the axes of the image array.
            Currently supported shape_orders are:
            - "HWC"
            - "CHW"
        with_meta (bool): Whether a simple numpy.ndarray is returned or a namedtuple that contains the image
            and it's meta info is returned.
            - If True, transformed image array with meta-info (ImageWithMeta) is returned
            - If False, transformed image array (numpy.ndarray) is returned
    """
    assert ((image is None and image_with_meta is not None) or
            (image is not None and image_with_meta is None)), "Both image and image_with_meta cannot be None or not None simultaneously."

    assert isinstance(image, (type(None), np.ndarray, np.generic)), (
        "image must be of type NoneType, np.ndarray or np.generic and not {}".format(type(image)))

    assert isinstance(image_with_meta, (type(None), ImageWithMeta)), (
        "image must be of type NoneType or ImageWithMeta and not {}".format(type(image_with_meta)))

    ############################################################################
    # META TYPE
    ############################################################################
    if meta_type:
        scale = meta_type.scale
        channel_order = meta_type.channel_order
        shape_order = meta_type.shape_order
        dtype = meta_type.dtype

    ############################################################################
    # IMAGE WITH META
    ############################################################################
    if image_with_meta is not None:
        logger.debug("TRANSFORM IMAGE: scale {} to {} ::: channel_order {} to {}"
                     .format(image_with_meta.scale, scale, image_with_meta.channel_order, channel_order))
        image = image_with_meta.image

        if (image_with_meta.scale != scale) and scale is not None:
            if scale == 1:
                image = (image / 255.0)  # side effect is that it converts uint8 to float32
            elif scale == 255:
                image = image * 255.0
            else:
                raise ValueError("scale must be either `1` or `255`, instead got {}".format(scale))

        if (image_with_meta.channel_order != channel_order) and channel_order is not None:
            # NOTE: Using `image = image[..., ::-1]` instead of cv2.COLOR_BGR2RGB because it doesn't work with float32.
            if channel_order == "RGB":
                image = image[..., ::-1]
            elif channel_order == "BGR":
                image = image[..., ::-1]
            else:
                raise ValueError("channel_order must be either 'RGB' or 'BGR', instead got {}".format(channel_order))

        if (image_with_meta.shape_order != shape_order) and shape_order is not None:
            # TODO: Swap axes
            raise NotImplementedError

        if (image_with_meta.dtype != dtype) and dtype is not None:
            image = image.astype(dtype)

        if with_meta is True:
            image = ImageWithMeta(image=image, scale=scale, channel_order=channel_order,
                                  shape_order=shape_order, dtype=dtype)

    ############################################################################
    # IMAGE
    ############################################################################
    else:
        if channel_order == "REVERSE_CHANNELS":
            # Make channels [0, 1, 2] into [2, 1, 0]
            image = image[..., ::-1]

    return image


def transform_image_batch(image_batch, input_meta_type, output_meta_type):
    """."""
    output_batch = np.empty(image_batch.shape, dtype=output_meta_type.dtype)
    for i in range(len(image_batch)):
        image_with_meta = add_meta_to_image(image_batch[i], meta_type=input_meta_type)
        output_batch[i] = transform_image(image_with_meta=image_with_meta, meta_type=output_meta_type, with_meta=False)
    return output_batch



"""Database Functions."""
from .models import Job, Task, db


#######################################
# Job Stuff
#######################################
def insert_new_job(uuid=None, name=None, payload_text=None, status='PENDING'):
    """Insert a new job and return it.

    Creates the job attributes if the attributes are NOT None.

    Args:
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.
    """
    # db = get_db()

    job_kwargs = {}
    if uuid is not None:
        job_kwargs['uuid'] = uuid
    if name is not None:
        job_kwargs['name'] = name
    if payload_text is not None:
        job_kwargs['payload_text'] = payload_text
    if status is not None:
        job_kwargs['status'] = status

    new_job = Job(**job_kwargs)
    db.session.add(new_job)
    db.session.commit()
    db.session.flush()
    return new_job


def select_job_all():
    """Select all jobs."""
    # db = get_db()
    all_jobs = db.session.query(Job).all()
    return all_jobs


def select_job_from_id(job_id):
    """Select job from ID."""
    # db = get_db()
    job = (db.session.query(Job)
           .filter_by(id=job_id)
           .first())
    return job


def select_job_uuid(job_uuid):
    """Select job from UUID."""
    # db = get_db()
    job = (db.session.query(Job)
           .filter_by(uuid=job_uuid)
           .first())
    return job


def update_job(job_id, uuid=None, name=None, payload_text=None, status=None):
    """Update job.

    Updates the job attributes if the attributes are NOT None.

    Args:
        job_id (int): The ID under which to create a task.
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.
    """
    # db = get_db()

    # Get job from job_id
    if job_id is None:
        raise AttributeError("job_id cannot be None")

    job = select_job_from_id(job_id)

    # Update job
    if uuid is not None:
        job.uuid = uuid
    if name is not None:
        job.name = name
    if payload_text is not None:
        job.payload_text = payload_text
    if status is not None:
        job.status = status

    db.session.add(job)
    db.session.commit()
    db.session.flush()
    return job


#######################################
# Task Stuff
#######################################
def insert_new_task(job_id, uuid=None, name=None, payload_text=None, status='PENDING'):
    """Insert a new task and return it.

    Creates the task attributes if the attributes are NOT None.

    Args:
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.
    """
    # db = get_db()

    task_kwargs = {'job_id': job_id}

    if uuid is not None:
        task_kwargs['uuid'] = uuid
    if name is not None:
        task_kwargs['name'] = name
    if payload_text is not None:
        task_kwargs['payload_text'] = payload_text
    if status is not None:
        task_kwargs['status'] = status

    new_task = Task(**task_kwargs)
    db.session.add(new_task)
    db.session.commit()
    db.session.flush()
    return new_task


def select_task_from_id(task_id):
    """Select task from id."""
    # db = get_db()
    task = (db.session.query(Task)
            .filter_by(id=task_id)
            .first())
    return task


def select_task_from_uuid(task_uuid):
    """Select task from UUID."""
    # db = get_db()
    task = (db.session.query(Task)
            .filter_by(uuid=task_uuid)
            .first())
    return task


def update_task(task_id, uuid=None, name=None, payload_text=None, status=None, update_parent_job=None):
    """Update task.

    Updates the task attributes if the attributes are NOT None.

    Args:
        task_id (int): The ID under which to create a task.
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.
    """
    # db = get_db()

    # Get task from task_id
    if task_id is None:
        raise AttributeError("task_id cannot be None")

    task = select_task_from_id(task_id)

    # Update task
    if uuid is not None:
        task.uuid = uuid
    if name is not None:
        task.name = name
    if payload_text is not None:
        task.payload_text = payload_text
    if status is not None:
        task.status = status

    db.session.add(task)
    db.session.commit()
    db.session.flush()

    if update_parent_job == "on_first_task":
        job = select_job_from_id(task.job_id)
        # Check if this task is the first task of the job.
        if job.tasks[0].id == task.id:
            update_job(job.id, status=task.status)
    elif update_parent_job == "on_last_task":
        job = select_job_from_id(task.job_id)
        # Check if this task is the last task of the job.
        if job.tasks[-1].id == task.id:
            update_job(job.id, status=task.status)
    elif update_parent_job == "on_this_task":
        job = select_job_from_id(task.job_id)
        update_job(job.id, status=task.status)
    elif not update_parent_job:
        pass  # Don't update in this case.
    else:
        raise ValueError("update_parent_job must be 'on_first_task', 'on_last_task', 'on_this_task' or Null, not {}."
                         .format(update_parent_job))

    return task



"""Utility functions."""
import numpy as np


def caffe_load_image(image_filename):
    """Load image using caffe.io.load_image.

    This is to maintain shape expectation across the caffe library.

    Args:
        image_filename (str): String filename.

    Returns:
        numpy.ndarray: an image with the following properties:
            shape: [Height, Width, Channels]
            channel_order: RGB
            scale: [0, 1]
            dtype: np.float32
    """
    import caffe
    return caffe.io.load_image(image_filename, color=True)


def caffe_load_image_batch(image_filenames, batch_size=None):
    """Load image using caffe.io.load_image.

    This is to maintain shape expectation across the caffe library.

    Args:
        image_filename (list of str): List of string filenames.
        batch_size (int): If batch_size is None, then all filenames are read.
            Otherwise only the first `batch_size` number of filenames are read.

    Returns:
        numpy.ndarray: an image with the following properties:
            shape: [batch_size, Height, Width, Channels]
            channel_order: RGB
            scale: [0, 1]
            dtype: np.float32
    """
    if batch_size is None:
        batch_size = len(image_filenames)

    image_batch = [caffe_load_image(image_filename) for image_filename in image_filenames[:batch_size]]
    image_batch = np.array(image_batch)  # converting list into numpy array

    return image_batch


# TODO (nitred): LRU cache
def get_caffe_transformer(net_input_shape, mean_bgr_255=None):
    """Transform a batch of images which were loaded by caffe.io.load_image.

    Transformations:
        - mean subtraction (if mean provided)
        - transposes data to become [Channels x Height x Width]
        - swaps channels to convert RGB to BGR
        - scales the data to [0., 255.]

    Args:
        net_input_shape (numpy.ndarray): The expected 4-dimensional shape of the network.
            The first dimension i.e. the batch_size doesn't really matter.
            Usually the expected shape is [BATCH_SIZE, Height, Width, Channels]
        mean_bgr_255 (numpy.ndarray): 1-dimensional array of means.
            Channel order should be BGR and scale should be [0., 255.]

    Returns:
        caffe.io.Transformer: With all standard transformations set.
    """
    import caffe
    transformer = caffe.io.Transformer({'data': net_input_shape})
    if mean_bgr_255 is not None:
        transformer.set_mean('data', mean_bgr_255)
    transformer.set_transpose('data', (2, 0, 1))
    transformer.set_channel_swap('data', (2, 1, 0))
    transformer.set_raw_scale('data', 255.0)
    return transformer


def caffe_transform_batch(X, net_input_shape, mean_bgr_255=None):
    """Transform a batch of images which were loaded by caffe.io.load_image.

    Transformations:
        - mean subtraction (if mean provided)
        - transposes data to become [Channels x Height x Width]
        - swaps channels to convert RGB to BGR
        - scales the data to [0., 255.]

    Args:
        X (numpy.ndarray):  A batch of images of shape [BATCH_SIZE, Height, Width, RGB-Channels].
            Can be obtained by using `caffe_utils.caffe_load_image`.
    """
    transformer = get_caffe_transformer(net_input_shape, mean_bgr_255)
    transformed_batch = np.array([transformer.preprocess('data', image) for image in X])
    return transformed_batch


def caffe_load_network_with_input_batch(net, X, mean_bgr_255=None, net_input_blob_name='data'):
    """Load the network with the input batch `inplace`.

    Args:
        net (caffe.Network): The network to load the input batch.
        X (numpy.ndarray):  A batch of images of shape [BATCH_SIZE, Height, Width, RGB-Channels].
            Can be obtained by using `caffe_utils.caffe_load_image`.
        mean_bgr_255 (numpy.ndarray): 1-dimensional array of means.
            Channel order should be BGR and scale should be [0., 255.]
        net_input_blob_name (str): The input blob name of the network. Default blob name is "data".

    Returns:
        net: The network is loaded with input inplace but it's returned anyway.
    """
    net_input_shape = net.blobs[net_input_blob_name].data.shape
    transformed_batch = caffe_transform_batch(X, net_input_shape, mean_bgr_255)
    net.blobs[net_input_blob_name].reshape(*transformed_batch.shape)
    net.blobs[net_input_blob_name].data[...] = transformed_batch
    return net



"""Useful pickling functions.

WARNING! Remember to use `from __future__ import absolute_import`
to avoid naming conflicts with the `pickle` module from standard library.
"""
from __future__ import absolute_import

try:
    import cPickle as pickle
except Exception as ex:
    import pickle


def write_pickle(obj, filename, encoding='ASCII'):
    """Write an object to file as a pickle.

    No error handling is done by this function!

    Args:
        obj (any): Any serializable object.
        filename (str): Full path of filename, ideally to have `.pkl` extention.

    Returns:
        None
    """
    pickle.dump(obj,
                open(filename, 'wb'),
                protocol=pickle.HIGHEST_PROTOCOL)


def read_pickle(filename, encoding='ASCII'):
    """Read an object from a pickle file.

    No error handling is done by this function!

    Args:
        filename (str): Full path of filename.

    Returns:
        any (obj): Object within the file. It can be of any type.
    """
    with open(filename, 'rb') as f:
        obj = pickle.load(f, encoding=encoding)

    return obj



"""Basic Image Utilities."""
import math

import cv2

import numpy as np


def image_array_to_image_batch(image):
    """Convert image_array to image_batch.

    Args:
        image (ndarray): 3D image array either HWC or CHW

    Returns:
        ndarray: 4D image_batch with batch_size being 1 and NHWC or NCHW format.
    """
    assert image.ndim == 3, "image must be 3D but instead was {}D".format(image.ndim)
    return image[None, :]


def resize_image_if_exceeds_max_area(image, max_area):
    """Resize image if it exceeds max area while maintaining aspect ratio.

    Source: https://stackoverflow.com/questions/33701929/how-to-resize-an-image-in-python-while-retaining-aspect-ratio-given-a-target-s/33702454

    Args:
        image (numpy.ndarray): An image array of HWC shape order and np.uint8 dtype.
        max_area (int): The number of pixels within the image i.e. height * width.

    Returns:
        numpy.ndarray: Resized image maintaining aspect ratio if image_area > max_area.
    """
    img_height, img_width = image.shape[0], image.shape[1]
    img_area = img_height * img_width
    if img_area > max_area:
        aspect_ratio = img_width / img_height
        new_height = int(math.sqrt(max_area / aspect_ratio) + 0.5)
        new_width = int((new_height * aspect_ratio) + 0.5)
        image = cv2.resize(image, (new_width, new_height))
    return image


def img_rotation(cv2_image, rg, fill_mode="constant", cval=0.):
    """."""
    angle = rg
    (h, w) = cv2_image.shape[: 2]
    img_center = (w / 2, h / 2)
    rot_matrix = cv2.getRotationMatrix2D(img_center, angle, 1.0)
    cv2_image = cv2.warpAffine(cv2_image, rot_matrix, (w, h), borderMode=cv2.BORDER_REPLICATE)

    return cv2_image


def img_shift(cv2_image, wrg, hrg, fill_mode="constant", cval=0.):
    """."""
    (h, w) = cv2_image.shape[: 2]

    t_w = wrg * w
    t_h = hrg * h
    t_w = int(t_w)
    t_h = int(t_h)
    transl_matrix = np.float32([[1, 0, t_w], [0, 1, t_h]])
    cv2_image = cv2.warpAffine(cv2_image, transl_matrix, (w, h), borderMode=cv2.BORDER_REPLICATE)

    return cv2_image


def img_horizontal_flip(x):
    """."""
    for i in range(x.shape[2]):
        x[:, :, i] = np.fliplr(x[:, :, i])

    return x


def img_scale(cv2_image, max_scale):
    """."""
    r_scale = max_scale
    r_height = int(np.round(cv2_image.shape[0] * (1 + r_scale)))
    r_width = int(np.round(cv2_image.shape[1] * (1 + r_scale)))
    if r_height <= cv2_image.shape[0]:
        cv2_cropped_image = cv2_image[int((cv2_image.shape[0] - r_height) / 2):
                                      int((cv2_image.shape[0] - r_height) / 2 + r_height),
                                      int((cv2_image.shape[1] - r_width) / 2):
                                      int((cv2_image.shape[1] - r_width) / 2 + r_width)]
        cv2_cropped_image = cv2.resize(cv2_cropped_image, (cv2_image.shape[0], cv2_image.shape[1]))
    else:
        cv2_cropped_image = np.zeros((r_height, r_width, 3), dtype=np.uint8)
        cv2_cropped_image = cv2.copyMakeBorder(cv2_image,
                                               top=int((r_height - cv2_image.shape[0]) / 2),
                                               bottom=int((r_height - cv2_image.shape[0]) / 2),
                                               left=int((r_width - cv2_image.shape[1]) / 2),
                                               right=int((r_width - cv2_image.shape[1]) / 2),
                                               borderType=cv2.BORDER_REPLICATE)
        cv2_cropped_image = cv2.resize(cv2_cropped_image, (cv2_image.shape[0], cv2_image.shape[1]))

    return cv2_cropped_image



"""An python api for the endpoints in job_status."""
import logging

import requests

from .utils import DotMapJob, DotMapTask

logger = logging.getLogger(__name__)


# ##############################################################################
# ############################ CREATE ##########################################
# ##############################################################################
def create_job(uuid=None, name=None, payload_text=None, status=None):
    """API for /create_job endpoint. Creates a new job and returns a job object.

    Args:
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.

    Returns:
        DotMap (job): A DotMap object that contains the details job and its tasks.
    """
    logger.debug('web_api/create_job: name: {}, uuid: {}'.format(name, uuid))
    response = requests.post(url="http://localhost:27131/job_status/create_job",
                             data={
                                 'uuid': uuid,
                                 'name': name,
                                 'payload_text': payload_text,
                                 'status': status,
                             })
    job_json = response.json()
    job = DotMapJob(job_json)
    return job


def create_task(job_id, uuid=None, name=None, payload_text=None, status=None):
    """API for /create_task endpoint. Creates a new task and returns a task object.

    Args:
        job_id (int): The ID under which to create a task.
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.

    Returns:
        DotMap (Task): A DotMap object that contains the details of the task.
    """
    logger.debug('web_apli/create_task: job_id: {}'.format(job_id))
    response = requests.post(url="http://localhost:27131/job_status/create_task",
                             data={
                                 'job_id': job_id,
                                 'uuid': uuid,
                                 'name': name,
                                 'payload_text': payload_text,
                                 'status': status,
                             })
    task_json = response.json()
    task = DotMapTask(task_json)
    return task


# ##############################################################################
# ############################# UPDATE #########################################
# ##############################################################################
def update_job(job_id, uuid=None, name=None, payload_text=None, status=None):
    """API for /update_job endpoint. Updates job.

    Args:
        job_id (int): The id of the job which need to be updated.
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.

    Returns:
        DotMap (Task): A DotMap object that contains the details of the job.
    """
    logger.debug('web_apli/update_job: job_id: {}, uuid: {}, name: {}, status: {}, payload_text: {}'
                 .format(job_id, uuid, name, status, payload_text))
    response = requests.post(url="http://localhost:27131/job_status/update_job",
                             data={
                                 'job_id': job_id,
                                 'uuid': uuid,
                                 'name': name,
                                 'payload_text': payload_text,
                                 'status': status,
                             })
    job_json = response.json()
    job = DotMapJob(job_json)
    return job


def update_task(task_id, uuid=None, name=None, payload_text=None, status=None, update_parent_job=None):
    """API for /update_task endpoint. Updates task.

    Args:
        task_id (int): The id of the task which need to be updated.
        uuid (str): Unique custom ID.
        name (str): Name.
        payload_text (str): Text payload.
        status (str): Status.

    Returns:
        DotMap (Task): A DotMap object that contains the details of the task.
    """
    logger.debug('web_api/update_task: task_id: {}, uuid: {}, name: {}, status: {}, payload_text: {}, update_parent_job: {}'
                 .format(task_id, uuid, name, status, payload_text, update_parent_job))
    response = requests.post(url="http://localhost:27131/job_status/update_task",
                             data={
                                 'task_id': task_id,
                                 'uuid': uuid,
                                 'name': name,
                                 'payload_text': payload_text,
                                 'status': status,
                                 'update_parent_job': update_parent_job
                             })
    task_json = response.json()
    task = DotMapTask(task_json)
    return task


# ##############################################################################
# ############################# READ ###########################################
# ##############################################################################
def read_job(job_id):
    """API for /read_job endpoint. Read job.

    Args:
        job_id (int): The id of the job to read.

    Returns:
        DotMap (job): A DotMap object that contains the details job and its tasks.
    """
    logger.debug('web_api/read_job: job_id: {}'.format(job_id))
    response = requests.get(url="http://localhost:27131/job_status/read_job", params={'job_id': job_id})
    job_dict = response.json()
    job = DotMapJob(job_dict)
    return job


def read_all_jobs():
    """API for /read_job endpoint. Read job.

    Returns:
        DotMap (list of jobs): A DotMap object that contains the details job and its tasks.
    """
    logger.debug('web_api/read_all_jobs')
    response = requests.get(url="http://localhost:27131/job_status/read_all_jobs")
    jobs = response.json()
    jobs = [DotMapJob(job_dict) for job_dict in jobs]
    return jobs


def read_task(task_id):
    """API for /read_task endpoint. Read task.

    Args:
        task_id (int): The id of the task to read.

    Returns:
        DotMap (Task): A DotMap object that contains the details of the task.
    """
    logger.debug('web_api/read_task: task_id: {}'.format(task_id))
    response = requests.get(url="http://localhost:27131/job_status/read_task",
                            params={'task_id': task_id})
    task_dict = response.json()
    task = DotMapTask(task_dict)
    return task


def read_task_from_uuid(task_uuid):
    """API for /read_task endpoint. Read task.

    Args:
        task_uuid (str): The uuid of the task to read.

    Returns:
        DotMap (Task): A DotMap object that contains the details of the task.
    """
    logger.debug('web_api/read_task_from_uuid: task_uuid: {}'.format(task_uuid))
    response = requests.get(url="http://localhost:27131/job_status/read_task_from_uuid",
                            params={'task_uuid': task_uuid})
    task_dict = response.json()
    task = DotMapTask(task_dict)
    return task


if __name__ == "__main__":
    job = create_job()
    print(job.as_json())

    task = create_task(job_id=job.id)
    print(task.as_json())

    task = update_task(task_id=task.id, status='NOT PENDING')
    print(task.as_json())

    old_job = read_job(job_id=job.id)
    print(old_job.as_json())

    old_task = read_task(task_id=task.id)
    print(old_task.as_json())



"""Config read and write.

################################################################################
# Modules required
################################################################################
* config_from_path
  - ruamel.yaml
  - trafaret_config
* read_config
  - pyyaml
"""
import os
import sys


def config_from_path(filepath, config_schema=None):
    """Read a config file from a given path.

    Args:
        filepath (str)  :A string representing the full path of your configfile. eg. /mnt/data/myconfig.yaml
        config_schema (trafaret): Trafaret object that defines the schema of the config.
            If None, then trafaret validation is not used.
            Example:
            ```
            import trafaret as tr
            config_schema = tr.Dict({
                tr.Key('project_name'):
                    tr.Dict({
                        'db_path': tr.String(),
                        'username': tr.String(),
                        'password': tr.String(),
                    }),
            })
            ```
            Trafaret docs: http://trafaret.readthedocs.io/en/latest/
    Return:
        config json
    """
    import ruamel.yaml
    from trafaret_config import read_and_validate
    if config_schema is None:
        with open(filepath) as stream:
            config = ruamel.yaml.load(stream, ruamel.yaml.RoundTripLoader)
            return config
    else:
        config = read_and_validate(filepath, config_schema)
        return config


def config_from_env(key, config_schema=None):
    """Read config from a file path in os.env.

    Args:
        key (str)   : Key represents an evironment variable to read config path
        config_schema (trafaret): Trafaret object that defines the schema of the config.
            If None, then trafaret validation is not used.
            Example:
            ```
            import trafaret as tr
            config_schema = tr.Dict({
                tr.Key('project_name'):
                    tr.Dict({
                        'db_path': tr.String(),
                        'username': tr.String(),
                        'password': tr.String(),
                    }),
            })
            ```
            Trafaret docs: http://trafaret.readthedocs.io/en/latest/
    Return:
        config json
    """
    filepath = os.getenv(key, default=None)
    if not filepath:
        sys.stderr.write("Passed key does not exist: {0}".format(key))
        raise AttributeError('Key {} does not exist in environment.'.format(key))

    return config_from_path(filepath, config_schema)


def read_config(config_filename):
    """Read and return config from filename.

    Args:
        config_filename (str): Path to config file
    Returns:
        dict: Config file as a dictionary.
    """
    import yaml
    with open(config_filename, 'r') as f:
        config = yaml.load(f)
    return config



import re
from codecs import open  # To use a consistent encoding
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


# Get version without importing, which avoids dependency issues
def get_version():
    with open('nr_common/__init__.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                         version_file.read()).group('version')


install_requires = ['future', 'ruamel.yaml>=0.15.35', 'trafaret_config>=1.0.1', 'pyyaml>=3.12']

test_requires = ['pytest', 'pytest-sugar', 'pytest-asyncio', 'pytest-cov', ]


setup(
    name='nr-common',
    description="Common python functionalities aimed to be at least compatible with Python3.",
    long_description=long_description,
    version=get_version(),
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=['pytest-runner'],
    tests_require=test_requires,
    packages=find_packages(),
    zip_safe=False,
    author="Nitish Reddy Koripalli",
    url="https://github.com/nitred/nr-common",
    download_url="https://github.com/nitred/nr-common/archive/{}.tar.gz".format(get_version()),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6", ]
)



"""Docstring for module."""
from flask_sqlalchemy import SQLAlchemy

# def get_db():
#     """Get database object from the Flask globals.
#
#     NOTE: This is an expectation of every Flask App that uses this blueprint to
#         set the `_db` attribute as the Flask-SQLAlchemy object.
#     """
#     db = current_app._db
#
#     if db is None:
#         raise AttributeError("Job Status blueprint needs `_db` attribute set in `flask.current_app`.")
#     else:
#         return db
#
#
# db = get_db()

db = SQLAlchemy()


class Base(db.Model):
    """Base class for all the tables.

    Consists of two default columns `created_at` and `modified_at` .
    """

    __abstract__ = True
    id = db.Column(db.Integer, primary_key=True)
    created_at = db.Column(db.DateTime, default=db.func.current_timestamp())
    modified_at = db.Column(db.DateTime,
                            default=db.func.current_timestamp(),
                            onupdate=db.func.current_timestamp())


class Job(Base):
    """A generic Job db data structure that has sub-tasks."""

    __tablename__ = 'job'

    tasks = db.relationship("Task", backref="job")

    uuid = db.Column(db.String(120), nullable=True, index=True)
    name = db.Column(db.String(256), nullable=True)
    payload_text = db.Column(db.Text(), nullable=True)
    status = db.Column(db.String(120), nullable=True)

    def __repr__(self):
        """String representation of the class."""
        return '<Job %r>' % self.id

    def as_json(self):
        """."""
        json_repr = {
            'id': self.id,
            'uuid': self.uuid,
            'name': self.name,
            'status': self.status,
            'payload_text': self.payload_text,
            'tasks': [task.as_json() for task in self.tasks]
        }
        return json_repr


class Task(Base):
    """A task that is a sub-task of a Job."""

    __tablename__ = 'task'

    # 'job' from backref
    job_id = db.Column(db.Integer, db.ForeignKey('job.id'))

    uuid = db.Column(db.String(120), nullable=True, index=True)
    name = db.Column(db.String(256), nullable=True)
    payload_text = db.Column(db.Text())
    status = db.Column(db.String(120), nullable=True)

    def __repr__(self):
        """String representation of the class."""
        return '<Task %r>' % self.id

    def as_json(self):
        """."""
        json_repr = {
            'job_id': self.job_id,
            'id': self.id,
            'uuid': self.uuid,
            'name': self.name,
            'payload_text': self.payload_text,
            'status': self.status,
        }
        return json_repr



"""Multiprocessing functions."""
from functools import wraps


def mproc_func(func):
    """Decorate functions to collect results into mproc_results_queue.

    Args:
        mproc_results_queue (mp.Manager.Queue): Queue to put results into.
        mproc_call_index (int): Index of the function call in the mproc_async.
    """
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        mproc_results_queue = kwargs.pop('mproc_results_queue')
        mproc_call_index = kwargs.pop('mproc_call_index')
        # Pass args and kwargs as is to the function.
        try:
            result = func(*args, **kwargs)
        except Exception as ex:
            print("EXCEPTION: {}".format(args))
            print(ex.args)
            result = None
        # Put result into result_queue.
        mproc_results_queue.put((mproc_call_index, result))

    return func_wrapper


def mproc_async(async_func, iterof_iterof_args=None, n_processes=1, queue_size=1):
    """Yield results of a async_function called multiple times using mproc.

    Args:
        async_func (func): A function that is called multiple times using mproc.
            The function MUST CONTAIN AN ARGUMENT called `result_queue`.
            OR
            A function that returns results and has a `mproc_func` DECORATOR.
        iterof_iterof_args (iter of iter): The iterable of arguments that must be passed
            to the `async_func`. The size of the iterable defines the number of mproc
            async function calls.

            Case 1: iterof_iterof_args: [[1], [2], [3]], then calls are f(1), f(2), f(3).

            Case 2: iterof_iterof_args: [[1,2,3], [4,5,6]], then calls are f(1,2,3), f(4,5,6).

            Case 3: iterof_iterof_args: [[], [], []], then calls are f(), f(), f().

        n_processes (int): Number of processes in pool.
        queue_size (int): The size of the internal queue that is used to collect
            results. This helps control memory usage.

    Returns:
        generator: A generator of (call_index, result) of async_func calls.
    """
    # Allocate Multiprocessing Resources
    import multiprocessing as mp
    pool = mp.Pool(n_processes)
    manager = mp.Manager()
    results_queue = manager.Queue(queue_size)
    print("MPROC          : Allocated resources")

    # Produce Async Task Processes
    func_calls = 0
    for call_index, args in enumerate(iterof_iterof_args):
        pool.apply_async(async_func, args=(*args,), kwds={'mproc_results_queue': results_queue,
                                                          'mproc_call_index': call_index})
        func_calls += 1
    else:
        print("MPROC          : Applied {} async calls".format(func_calls))

    # Consume results from queue
    item_count = 0
    while True:
        yield results_queue.get(block=True)
        results_queue.task_done()
        item_count += 1

        if item_count == func_calls:
            print("MPROC          : Joining queue, all calls completed.")
            results_queue.join()
            # Break generator loop
            break
